@@ -29,6 +29,14 @@
 //!
 //! Additionally, one can explicitly disable hot-reloading for a type.
 
+#[cfg(feature = "ab_glyph")]
+mod fonts;
+#[cfg(feature = "gltf")]
+mod gltf;
+
+#[cfg(test)]
+mod tests;
+
 pub use crate::dirs::DirLoadable;
 
 #[allow(unused)]
@@ -52,6 +60,9 @@ use serde::{Deserialize, Serialize};
 #[allow(unused)]
 use std::{borrow::Cow, io, sync::Arc};
 
+#[cfg(feature = "gltf")]
+pub use gltf::Gltf;
+
 /// An asset is a type loadable from raw bytes.
 ///
 /// `Asset`s can be loaded and retrieved by an [`AssetCache`].
@@ -113,9 +124,8 @@ pub trait Asset: Sized + Send + Sync + 'static {
     ///
     /// If `EXTENSION` is provided, you don't have to set this constant.
     ///
-    /// If this array is empty, loading an asset of this type returns
-    /// [`Error::NoDefaultValue`] unless a default value is provided with the
-    /// `default_value` method.
+    /// If this array is empty, loading an asset of this type returns an error
+    /// unless a default value is provided with the `default_value` method.
     const EXTENSIONS: &'static [&'static str] = &[Self::EXTENSION];
 
     /// Specifies a way to convert raw bytes into the asset.
@@ -211,39 +221,36 @@ pub trait Compound: Sized + Send + Sync + 'static {
     ///
     /// This function should not perform any kind of I/O: such concern should be
     /// delegated to [`Asset`]s.
-    fn load<S: Source>(cache: &AssetCache<S>, id: &str) -> Result<Self, Error>;
+    fn load<S: Source + ?Sized>(cache: &AssetCache<S>, id: &str) -> Result<Self, BoxedError>;
 
     /// Loads an asset and registers it for hot-reloading if necessary.
     ///
     /// This method is a internal implementation detail.
     #[doc(hidden)]
-    fn _load_and_record<S: Source, P: PrivateMarker>(
+    fn _load_and_record<S: Source + ?Sized, P: PrivateMarker>(
         cache: &AssetCache<S>,
         id: &SharedString,
     ) -> Result<Self, Error> {
         #[cfg(feature = "hot-reloading")]
-        {
-            use crate::utils::DepsRecord;
-
-            if Self::HOT_RELOADED {
-                let (asset, deps) = cache.record_load(id)?;
-                cache
-                    .source()
-                    ._add_compound::<Self, P>(id, DepsRecord(deps));
-                Ok(asset)
-            } else {
-                cache.no_record(|| Self::load(cache, id))
-            }
-        }
+        let res = if Self::HOT_RELOADED {
+            cache.record_load(id).map(|(asset, deps)| {
+                if let Some(reloader) = &cache.reloader {
+                    reloader.add_compound::<Self>(id.clone(), deps);
+                }
+                asset
+            })
+        } else {
+            cache.no_record(|| Self::load(cache, id))
+        };
 
         #[cfg(not(feature = "hot-reloading"))]
-        {
-            Self::load(cache, id)
-        }
+        let res = Self::load(cache, id);
+
+        res.map_err(|err| Error::new(id.clone(), err))
     }
 
     #[doc(hidden)]
-    fn _load_and_record_entry<S: Source, P: PrivateMarker>(
+    fn _load_and_record_entry<S: Source + ?Sized, P: PrivateMarker>(
         cache: &AssetCache<S>,
         id: SharedString,
     ) -> Result<CacheEntry, Error> {
@@ -255,6 +262,11 @@ pub trait Compound: Sized + Send + Sync + 'static {
     /// default). If so, you may want to implement [`NotHotReloaded`] for this
     /// type to enable additional functions.
     const HOT_RELOADED: bool = true;
+
+    #[doc(hidden)]
+    fn get_key<P: PrivateMarker>() -> Option<crate::key::AssetType> {
+        None
+    }
 }
 
 impl<A> Compound for A
@@ -262,34 +274,42 @@ where
     A: Asset,
 {
     #[inline]
-    fn load<S: Source>(cache: &AssetCache<S>, id: &str) -> Result<Self, Error> {
-        load_from_source(cache.source(), id)
+    fn load<S: Source + ?Sized>(cache: &AssetCache<S>, id: &str) -> Result<Self, BoxedError> {
+        Ok(load_from_source(cache.source(), id)?)
     }
 
     #[doc(hidden)]
-    fn _load_and_record<S: Source, P: PrivateMarker>(
+    fn _load_and_record<S: Source + ?Sized, P: PrivateMarker>(
         cache: &AssetCache<S>,
         id: &SharedString,
     ) -> Result<Self, Error> {
-        let asset = cache.no_record(|| Self::load(cache, id))?;
+        let asset = load_from_source(cache.source(), id)?;
 
         #[cfg(feature = "hot-reloading")]
         if A::HOT_RELOADED {
-            cache.source()._add_asset::<A, P>(id);
+            if let Some(reloader) = &cache.reloader {
+                reloader.add_asset::<Self>(id.clone());
+            }
         }
 
         Ok(asset)
     }
 
     const HOT_RELOADED: bool = Self::HOT_RELOADED;
+
+    #[doc(hidden)]
+    fn get_key<P: PrivateMarker>() -> Option<crate::key::AssetType> {
+        Some(crate::key::AssetType::of::<Self>())
+    }
 }
 
 impl<A> Compound for Arc<A>
 where
     A: Compound,
 {
-    fn load<S: Source>(cache: &AssetCache<S>, id: &str) -> Result<Self, Error> {
-        cache.load_owned::<A>(id).map(Arc::new)
+    fn load<S: Source + ?Sized>(cache: &AssetCache<S>, id: &str) -> Result<Self, BoxedError> {
+        let asset = cache.load_owned::<A>(id)?;
+        Ok(Arc::new(asset))
     }
 
     const HOT_RELOADED: bool = A::HOT_RELOADED;
@@ -345,6 +365,11 @@ pub trait Storable: Send + Sync + 'static {
     /// # Ok::<(), assets_manager::BoxedError>(())
     /// ```
     const _CHECK_NOT_HOT_RELOADED: () = [()][Self::HOT_RELOADED as usize];
+
+    #[doc(hidden)]
+    fn get_key<P: PrivateMarker>() -> Option<crate::key::AssetType> {
+        None
+    }
 }
 
 impl<A> Storable for A
@@ -447,6 +472,7 @@ macro_rules! serde_assets {
             }
 
             #[cfg(feature = $feature)]
+            #[cfg_attr(docsrs, doc(cfg(feature = $feature)))]
             impl<T> Asset for $name<T>
             where
                 T: for<'de> serde::Deserialize<'de> + Send + Sync + 'static,
@@ -526,6 +552,7 @@ macro_rules! image_assets {
             }
 
             #[cfg(feature = $feature)]
+            #[cfg_attr(docsrs, doc(cfg(feature = $feature)))]
             impl Asset for $name {
                 const EXTENSIONS: &'static [&'static str] = &[$( $ext ),*];
                 type Loader = loader::ImageLoader;
@@ -555,6 +582,13 @@ image_assets! {
         image::ImageFormat::Png,
         ["png"],
     );
+
+    /// An asset to load PNG images.
+    #[cfg(feature = "webp")]
+    struct Webp => (
+        image::ImageFormat::WebP,
+        ["webp"],
+    );
 }
 
 macro_rules! sound_assets {
@@ -586,6 +620,7 @@ macro_rules! sound_assets {
             }
 
             #[cfg(feature = $feature)]
+            #[cfg_attr(docsrs, doc(cfg(feature = $feature)))]
             impl Asset for $name {
                 const EXTENSIONS: &'static [&'static str] = &[$( $ext ),*];
                 type Loader = loader::SoundLoader;
