@@ -4,6 +4,7 @@ use crate::{
     asset::{DirLoadable, Storable},
     dirs::DirHandle,
     entry::{CacheEntry, CacheEntryInner},
+    error::ErrorKind,
     loader::Loader,
     source::{FileSystem, Source},
     utils::{BorrowedKey, HashMap, Key, OwnedKey, Private, RandomState, RwLock},
@@ -16,7 +17,7 @@ use crate::AssetGuard;
 use std::{any::TypeId, fmt, io, path::Path};
 
 #[cfg(feature = "hot-reloading")]
-use crate::utils::HashSet;
+use crate::{hot_reloading::HotReloader, utils::HashSet};
 
 #[cfg(feature = "hot-reloading")]
 use std::{cell::Cell, ptr::NonNull};
@@ -41,10 +42,7 @@ impl Map {
 
         let hash_builder = RandomState::new();
         let shards = (0..shards)
-            .map(|_| {
-                let map = HashMap::with_hasher(hash_builder.clone());
-                RwLock::new(map)
-            })
+            .map(|_| RwLock::new(HashMap::with_hasher(hash_builder.clone())))
             .collect();
 
         Map {
@@ -228,9 +226,12 @@ thread_local! {
 /// # }}
 /// # Ok::<(), Box<dyn std::error::Error>>(())
 /// ```
-pub struct AssetCache<S = FileSystem> {
-    source: S,
+pub struct AssetCache<S: ?Sized = FileSystem> {
+    #[cfg(feature = "hot-reloading")]
+    pub(crate) reloader: Option<HotReloader>,
+
     pub(crate) assets: Map,
+    source: S,
 }
 
 impl AssetCache<FileSystem> {
@@ -246,15 +247,39 @@ impl AssetCache<FileSystem> {
     }
 }
 
-impl<S> AssetCache<S> {
-    /// Creates a cache that loads assets from the given source.
+impl<S: Source> AssetCache<S> {
+    /// Creates a cache that loads assets from the given source and tries to
+    /// start hot-reloading (if feature `hot-reloading` is used).
+    ///
+    /// If hot-reloading fails to start, an error is logged.
     pub fn with_source(source: S) -> AssetCache<S> {
         AssetCache {
+            #[cfg(feature = "hot-reloading")]
+            reloader: HotReloader::make(&source),
+
             assets: Map::new(32),
             source,
         }
     }
+}
 
+impl<S: Source> AssetCache<S> {
+    /// Creates a cache that loads assets from the given source.
+    pub fn without_hot_reloading(source: S) -> AssetCache<S> {
+        AssetCache {
+            #[cfg(feature = "hot-reloading")]
+            reloader: None,
+
+            assets: Map::new(32),
+            source,
+        }
+    }
+}
+
+impl<S> AssetCache<S>
+where
+    S: ?Sized,
+{
     /// Returns a reference to the cache's [`Source`].
     #[inline]
     pub fn source(&self) -> &S {
@@ -266,7 +291,7 @@ impl<S> AssetCache<S> {
         RECORDING.with(|rec| {
             if let Some(mut recorder) = rec.get() {
                 let recorder = unsafe { recorder.as_mut() };
-                recorder.insert(self as *const Self as usize, key);
+                recorder.insert(self as *const Self as *const () as usize, key);
             }
         });
     }
@@ -316,6 +341,8 @@ impl<S> AssetCache<S> {
         type_id: TypeId,
         load: fn(&Self, SharedString) -> Result<CacheEntry, Error>,
     ) -> Result<CacheEntryInner, Error> {
+        log::trace!("Loading \"{}\"", id);
+
         let id = SharedString::from(id);
         let entry = load(self, id.clone())?;
         let key = OwnedKey::new_with(id, type_id);
@@ -391,7 +418,7 @@ impl<S> AssetCache<S> {
 
     /// Returns `true` if the cache contains the specified asset.
     #[inline]
-    pub fn contains<A: Compound>(&self, id: &str) -> bool {
+    pub fn contains<A: Storable>(&self, id: &str) -> bool {
         let key = BorrowedKey::new::<A>(id);
         self.assets.contains_key(key)
     }
@@ -421,20 +448,68 @@ impl<S> AssetCache<S> {
     pub fn contains_dir<A: DirLoadable>(&self, id: &str, recursive: bool) -> bool {
         self.get_cached_dir::<A>(id, recursive).is_some()
     }
+
+    /// Removes an asset from the cache, and returns whether it was present in
+    /// the cache.
+    ///
+    /// Note that you need a mutable reference to the cache, so you cannot have
+    /// any [`Handle`], [`AssetGuard`], etc when you call this function.
+    #[inline]
+    pub fn remove<A: Storable>(&mut self, id: &str) -> bool {
+        let key = BorrowedKey::new::<A>(id);
+
+        #[cfg(feature = "hot-reloading")]
+        if let Some(reloader) = &self.reloader {
+            reloader.remove_asset::<A>(SharedString::from(id));
+        }
+
+        self.assets.remove(key)
+    }
+
+    /// Takes ownership on a cached asset.
+    ///
+    /// The corresponding asset is removed from the cache.
+    #[inline]
+    pub fn take<A: Storable>(&mut self, id: &str) -> Option<A> {
+        let key = BorrowedKey::new::<A>(id);
+        self.assets.take(key).map(|e| {
+            let (asset, _id) = e.into_inner();
+
+            #[cfg(feature = "hot-reloading")]
+            if let Some(reloader) = &self.reloader {
+                reloader.remove_asset::<A>(_id);
+            }
+
+            asset
+        })
+    }
+
+    /// Clears the cache.
+    ///
+    /// Removes all cached assets and directories.
+    #[inline]
+    pub fn clear(&mut self) {
+        self.assets.clear();
+
+        #[cfg(feature = "hot-reloading")]
+        if let Some(reloader) = &self.reloader {
+            reloader.clear();
+        }
+    }
 }
 
 impl<S> AssetCache<S>
 where
-    S: Source,
+    S: Source + ?Sized,
 {
     #[cfg(feature = "hot-reloading")]
     pub(crate) fn record_load<A: Compound>(
         &self,
         id: &str,
-    ) -> Result<(A, HashSet<OwnedKey>), Error> {
-        let mut record = Record::new(self as *const Self as usize);
+    ) -> Result<(A, HashSet<OwnedKey>), crate::BoxedError> {
+        let mut record = Record::new(self as *const Self as *const () as usize);
 
-        let asset = if S::_support_hot_reloading::<Private>(&self.source) {
+        let asset = if self.reloader.is_some() {
             RECORDING.with(|rec| {
                 let old_rec = rec.replace(Some(NonNull::from(&mut record)));
                 let result = A::load(self, id);
@@ -484,14 +559,18 @@ where
     pub fn load_expect<A: Compound>(&self, id: &str) -> Handle<A> {
         #[cold]
         #[track_caller]
-        fn expect_failed(id: &str, err: Error) -> ! {
-            panic!("Failed to load essential asset \"{}\": {}", id, err)
+        fn expect_failed(err: Error) -> ! {
+            panic!(
+                "Failed to load essential asset \"{}\": {}",
+                err.id(),
+                err.reason()
+            )
         }
 
         // Do not use `unwrap_or_else` as closures do not have #[track_caller]
         match self.load(id) {
             Ok(h) => h,
-            Err(err) => expect_failed(id, err),
+            Err(err) => expect_failed(err),
         }
     }
 
@@ -537,50 +616,23 @@ where
     /// This can be useful if you need ownership on a non-clonable value.
     #[inline]
     pub fn load_owned<A: Compound>(&self, id: &str) -> Result<A, Error> {
+        let id = SharedString::from(id);
+        let asset = A::_load_and_record::<S, Private>(self, &id);
+
         #[cfg(feature = "hot-reloading")]
         if A::HOT_RELOADED && self.is_recording() {
-            let id = SharedString::from(id);
-            let key = OwnedKey::new::<A>(id.clone());
+            let key = OwnedKey::new::<A>(id);
             self.add_record(key);
-            return A::_load_and_record::<S, Private>(self, &id);
         }
 
-        A::load(self, id)
-    }
-
-    /// Removes an asset from the cache, and returns whether it was present in
-    /// the cache.
-    ///
-    /// Note that you need a mutable reference to the cache, so you cannot have
-    /// any [`Handle`], [`AssetGuard`], etc when you call this function.
-    #[inline]
-    pub fn remove<A: Compound>(&mut self, id: &str) -> bool {
-        let key = BorrowedKey::new::<A>(id);
-        self.assets.remove(key)
-    }
-
-    /// Takes ownership on a cached asset.
-    ///
-    /// The corresponding asset is removed from the cache.
-    #[inline]
-    pub fn take<A: Compound>(&mut self, id: &str) -> Option<A> {
-        let key = BorrowedKey::new::<A>(id);
-        self.assets.take(key).map(|e| e.into_inner())
-    }
-
-    /// Clears the cache.
-    ///
-    /// Removes all cached assets and directories.
-    #[inline]
-    pub fn clear(&mut self) {
-        self.assets.clear();
-
-        #[cfg(feature = "hot-reloading")]
-        self.source._clear::<Private>();
+        asset
     }
 }
 
-impl AssetCache<FileSystem> {
+impl<S> AssetCache<S>
+where
+    S: Source + Sync,
+{
     /// Reloads changed assets.
     ///
     /// This function is typically called within a loop.
@@ -600,7 +652,7 @@ impl AssetCache<FileSystem> {
     #[cfg_attr(docsrs, doc(cfg(feature = "hot-reloading")))]
     #[inline]
     pub fn hot_reload(&self) {
-        if let Some(reloader) = &self.source.reloader {
+        if let Some(reloader) = &self.reloader {
             reloader.reload(self);
         }
     }
@@ -622,13 +674,16 @@ impl AssetCache<FileSystem> {
     #[cfg_attr(docsrs, doc(cfg(feature = "hot-reloading")))]
     #[inline]
     pub fn enhance_hot_reloading(&'static self) {
-        if let Some(reloader) = &self.source.reloader {
+        if let Some(reloader) = &self.reloader {
             reloader.send_static(self);
         }
     }
 }
 
-impl<S> fmt::Debug for AssetCache<S> {
+impl<S> fmt::Debug for AssetCache<S>
+where
+    S: ?Sized,
+{
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         f.debug_struct("AssetCache")
             .field("assets", &self.assets)
@@ -637,21 +692,29 @@ impl<S> fmt::Debug for AssetCache<S> {
 }
 
 #[inline]
-fn load_single<A: Asset, S: Source>(source: &S, id: &str, ext: &str) -> Result<A, Error> {
+fn load_single<A, S>(source: &S, id: &str, ext: &str) -> Result<A, ErrorKind>
+where
+    A: Asset,
+    S: Source + ?Sized,
+{
     let content = source.read(id, ext)?;
     let asset = A::Loader::load(content, ext)?;
     Ok(asset)
 }
 
-pub(crate) fn load_from_source<A: Asset, S: Source>(source: &S, id: &str) -> Result<A, Error> {
-    let mut error = Error::NoDefaultValue;
+pub(crate) fn load_from_source<A, S>(source: &S, id: &str) -> Result<A, Error>
+where
+    A: Asset,
+    S: Source + ?Sized,
+{
+    let mut error = ErrorKind::NoDefaultValue;
 
     for ext in A::EXTENSIONS {
         match load_single(source, id, ext) {
             Err(err) => error = err.or(error),
-            asset => return asset,
+            Ok(asset) => return Ok(asset),
         }
     }
 
-    A::default_value(id, error)
+    A::default_value(id, Error::from_kind(id.into(), error))
 }
