@@ -1,6 +1,6 @@
 use crate::{
     source::{DirEntry, Source},
-    Asset, AssetCache, Compound, Error, Handle, SharedString,
+    Asset, AssetCache, BoxedError, Compound, Error, Handle, SharedString,
 };
 
 use std::{fmt, io, marker::PhantomData};
@@ -21,7 +21,7 @@ use std::{fmt, io, marker::PhantomData};
 /// ```no_run
 /// # cfg_if::cfg_if! { if #[cfg(all(feature = "json", feature = "flac"))] {
 /// use assets_manager::{
-///     Compound, Error, AssetCache, SharedString,
+///     Compound, BoxedError, AssetCache, SharedString,
 ///     asset::{DirLoadable, Json, Flac},
 ///     source::{DirEntry, Source},
 /// };
@@ -33,14 +33,14 @@ use std::{fmt, io, marker::PhantomData};
 ///
 /// // Specify how to load a playlist
 /// impl Compound for Playlist {
-///     fn load<S: Source>(cache: &AssetCache<S>, id: &str) -> Result<Self, Error> {
+///     fn load<S: Source + ?Sized>(cache: &AssetCache<S>, id: &str) -> Result<Self, BoxedError> {
 ///         // Read the manifest (a list of ids)
 ///         let manifest = cache.load::<Json<Vec<String>>>(id)?.read();
 ///
 ///         // Load each sound
 ///         let sounds = manifest.0.iter()
 ///             .map(|id| Ok(cache.load::<Flac>(id)?.cloned()))
-///             .collect::<Result<_, Error>>()?;
+///             .collect::<Result<_, BoxedError>>()?;
 ///
 ///         Ok(Playlist { sounds })
 ///     }
@@ -122,20 +122,14 @@ impl<A> Compound for CachedDir<A>
 where
     A: DirLoadable,
 {
-    fn load<S: Source>(cache: &AssetCache<S>, id: &str) -> Result<Self, Error> {
-        let mut ids = A::select_ids(cache.source(), id)?;
+    fn load<S: Source + ?Sized>(cache: &AssetCache<S>, id: &str) -> Result<Self, BoxedError> {
+        let mut ids =
+            A::select_ids(cache.source(), id).map_err(|err| Error::from_io(id.into(), err))?;
 
         // Remove duplicated entries
         ids.sort_unstable();
         ids.dedup();
 
-        // Pre-load inner assets
-        cache.no_record(|| {
-            for id in &ids {
-                let _ = cache.load::<A>(id);
-            }
-        });
-
         Ok(CachedDir {
             ids,
             _marker: PhantomData,
@@ -163,19 +157,22 @@ impl<A> Compound for CachedRecDir<A>
 where
     A: DirLoadable,
 {
-    fn load<S: Source>(cache: &AssetCache<S>, id: &str) -> Result<Self, Error> {
+    fn load<S: Source + ?Sized>(cache: &AssetCache<S>, id: &str) -> Result<Self, BoxedError> {
         // Load the current directory
         let this = cache.load::<CachedDir<A>>(id)?;
         let mut ids = this.get().ids.clone();
 
         // Recursively load child directories
-        cache.source().read_dir(id, &mut |entry| {
-            if let DirEntry::Directory(id) = entry {
-                if let Ok(child) = cache.load::<CachedRecDir<A>>(id) {
-                    ids.extend_from_slice(&child.get().ids);
+        cache
+            .source()
+            .read_dir(id, &mut |entry| {
+                if let DirEntry::Directory(id) = entry {
+                    if let Ok(child) = cache.load::<CachedRecDir<A>>(id) {
+                        ids.extend_from_slice(&child.get().ids);
+                    }
                 }
-            }
-        })?;
+            })
+            .map_err(|err| Error::from_io(id.into(), err))?;
 
         Ok(CachedRecDir {
             ids,
@@ -231,7 +228,7 @@ where
 /// A handle on a asset directory.
 ///
 /// This type provides methods to access assets within a directory.
-pub struct DirHandle<'a, A, S> {
+pub struct DirHandle<'a, A, S: ?Sized> {
     inner: DirHandleInner<'a, A>,
     cache: &'a AssetCache<S>,
 }
@@ -239,6 +236,7 @@ pub struct DirHandle<'a, A, S> {
 impl<'a, A, S> DirHandle<'a, A, S>
 where
     A: DirLoadable,
+    S: ?Sized,
 {
     #[inline]
     pub(crate) fn new(handle: Handle<'a, CachedDir<A>>, cache: &'a AssetCache<S>) -> Self {
@@ -280,7 +278,7 @@ where
 impl<'a, A, S> DirHandle<'a, A, S>
 where
     A: DirLoadable,
-    S: Source,
+    S: Source + ?Sized,
 {
     /// Returns an iterator over the assets in the directory.
     ///
@@ -295,17 +293,18 @@ where
     }
 }
 
-impl<A, S> Clone for DirHandle<'_, A, S> {
+impl<A, S: ?Sized> Clone for DirHandle<'_, A, S> {
     fn clone(&self) -> Self {
         *self
     }
 }
 
-impl<A, S> Copy for DirHandle<'_, A, S> {}
+impl<A, S: ?Sized> Copy for DirHandle<'_, A, S> {}
 
 impl<A, S> fmt::Debug for DirHandle<'_, A, S>
 where
     A: DirLoadable,
+    S: ?Sized,
 {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         f.debug_struct("DirHandle")
