@@ -10,8 +10,8 @@
 //! # Hot-reloading
 //!
 //! Hot-reloading enable assets to be reloaded automatically when the source it
-//! was loaded from was modified. It is only supported for the [`FileSystem`]
-//! source at the moment.
+//! was loaded from was modified. It requires the `Source` to support it. The
+//! built-in [`FileSystem`] source supports it out of the box.
 //!
 //! # Using a different source depending on the target platform
 //!
@@ -31,13 +31,14 @@
 //! # Ok::<(), std::io::Error>(())
 //! ```
 
-#[cfg(feature = "hot-reloading")]
-use crate::{utils::PrivateMarker, SharedString};
-
 use std::{borrow::Cow, io};
 
 #[cfg(doc)]
 use crate::{asset::DirLoadable, AssetCache};
+use crate::{
+    hot_reloading::{DynUpdateSender, EventSender},
+    BoxedError,
+};
 
 mod filesystem;
 pub use filesystem::FileSystem;
@@ -202,40 +203,24 @@ pub trait Source {
     /// ```
     fn exists(&self, entry: DirEntry) -> bool;
 
-    #[cfg(feature = "hot-reloading")]
-    #[doc(hidden)]
-    fn _add_asset<A: crate::Asset, P: PrivateMarker>(&self, _: &SharedString)
-    where
-        Self: Sized,
-    {
+    /// Returns a source to use with hot-reloading.
+    ///
+    /// This method returns `None` when the source does not support
+    /// hot-reloading. In this case, `configure_hot_reloading` will never be
+    /// called.
+    #[inline]
+    fn make_source(&self) -> Option<Box<dyn Source + Send>> {
+        None
     }
 
-    #[cfg(feature = "hot-reloading")]
-    #[doc(hidden)]
-    fn _clear<P: PrivateMarker>(&mut self)
-    where
-        Self: Sized,
-    {
-    }
-
-    #[cfg(feature = "hot-reloading")]
-    #[doc(hidden)]
-    fn _add_compound<A: crate::Compound, P: PrivateMarker>(
-        &self,
-        _: &SharedString,
-        _: crate::utils::DepsRecord,
-    ) where
-        Self: Sized,
-    {
-    }
-
-    #[cfg(feature = "hot-reloading")]
-    #[doc(hidden)]
-    fn _support_hot_reloading<P: PrivateMarker>(&self) -> bool
-    where
-        Self: Sized,
-    {
-        false
+    /// Configures hot-reloading.
+    ///
+    /// This method receives an `EventSender` to notify the hot-reloading
+    /// subsystem when assets should be reloaded. It returns a `DynUpdateSender`
+    /// to be notified of cache state changes.
+    #[inline]
+    fn configure_hot_reloading(&self, _events: EventSender) -> Result<DynUpdateSender, BoxedError> {
+        Err("this source does not support hot-reloading".into())
     }
 }
 
@@ -243,17 +228,90 @@ impl<S> Source for Box<S>
 where
     S: Source + ?Sized,
 {
+    #[inline]
     fn read(&self, id: &str, ext: &str) -> io::Result<Cow<[u8]>> {
         self.as_ref().read(id, ext)
     }
 
+    #[inline]
     fn read_dir(&self, id: &str, f: &mut dyn FnMut(DirEntry)) -> io::Result<()> {
         self.as_ref().read_dir(id, f)
     }
 
+    #[inline]
     fn exists(&self, entry: DirEntry) -> bool {
         self.as_ref().exists(entry)
     }
+
+    #[inline]
+    fn make_source(&self) -> Option<Box<dyn Source + Send>> {
+        self.as_ref().make_source()
+    }
+
+    #[inline]
+    fn configure_hot_reloading(&self, events: EventSender) -> Result<DynUpdateSender, BoxedError> {
+        self.as_ref().configure_hot_reloading(events)
+    }
+}
+
+impl<S> Source for &S
+where
+    S: Source + ?Sized,
+{
+    #[inline]
+    fn read(&self, id: &str, ext: &str) -> io::Result<Cow<[u8]>> {
+        (**self).read(id, ext)
+    }
+
+    #[inline]
+    fn read_dir(&self, id: &str, f: &mut dyn FnMut(DirEntry)) -> io::Result<()> {
+        (**self).read_dir(id, f)
+    }
+
+    #[inline]
+    fn exists(&self, entry: DirEntry) -> bool {
+        (**self).exists(entry)
+    }
+
+    #[inline]
+    fn make_source(&self) -> Option<Box<dyn Source + Send>> {
+        (**self).make_source()
+    }
+
+    #[inline]
+    fn configure_hot_reloading(&self, events: EventSender) -> Result<DynUpdateSender, BoxedError> {
+        (**self).configure_hot_reloading(events)
+    }
+}
+
+impl<S> Source for std::sync::Arc<S>
+where
+    S: Source + ?Sized,
+{
+    #[inline]
+    fn read(&self, id: &str, ext: &str) -> io::Result<Cow<[u8]>> {
+        self.as_ref().read(id, ext)
+    }
+
+    #[inline]
+    fn read_dir(&self, id: &str, f: &mut dyn FnMut(DirEntry)) -> io::Result<()> {
+        self.as_ref().read_dir(id, f)
+    }
+
+    #[inline]
+    fn exists(&self, entry: DirEntry) -> bool {
+        self.as_ref().exists(entry)
+    }
+
+    #[inline]
+    fn make_source(&self) -> Option<Box<dyn Source + Send>> {
+        (**self).make_source()
+    }
+
+    #[inline]
+    fn configure_hot_reloading(&self, events: EventSender) -> Result<DynUpdateSender, BoxedError> {
+        self.as_ref().configure_hot_reloading(events)
+    }
 }
 
 /// A [`Source`] that contains nothing.
